(****************************)
(*** Fonction utilitaires ***)
(****************************)
(** Conversion d'une liste en chaîne de caractères selon une fonction de transformation et un séparateur *)
let rec string_of_list conv sep l =
    match l with
      | [] -> ""
      | h :: [] -> (conv h)
      | h :: t -> ((conv h) ^ sep ^ (string_of_list conv sep t)) ;;

let string_of_list_delim conv sep delim1 delim2 l =
  delim1 ^ string_of_list conv sep l ^ delim2 ;;

(** Calcul de l'ensemble des parties *)
let powerset l =
  let rec ps lp l =
    match l with
      | [] -> [lp]
      | h :: t -> (ps lp t) @ (ps (lp @ [h]) t) in
  ps [] l ;;

(** Calcul de l'ensemble des parties ordonné *)
let powerset_sort l fcompare =
  let rec ps lp l fcompare =
    match l with
      | [] -> [List.sort fcompare lp]
      | h :: t -> (ps lp t fcompare) @ (ps (lp @ [h]) t fcompare) in
  ps [] l fcompare ;;

(** Récupération l'index d'un élément d'une liste *)
let index l a =
  let rec index_rec l a n =
    match l with
      | [] -> raise Not_found
      | h :: t -> if h = a then n else index_rec t a (n+1) in
  index_rec l a 0 ;;



(**************************)
(*** Opérateurs communs ***)
(**************************)
module Operators =
struct
(** Opérateurs *)
  type propop2 =
    | And | Or | Impl ;;
  type propop1 =
    | Neg ;;
  type propop0 =
    | True | False ;;
  type relop2 =
    | LE | LT | GE | GT | Eq | NEq ;;
  type exprop2 =
    | Plus | Minus ;;

(** Conversion des opérateurs en chaînes de caractères pour ASP *)
  let asp_of_relop2 = function
    | LE -> "<="
    | LT -> "<"
    | GE -> ">="
    | GT -> ">"
    | Eq -> "=="
    | NEq -> "!=" ;;
  let asp_of_exprop2 = function
    | Plus -> "+"
    | Minus -> "-" ;;

(** Conversion des opérateurs en chaînes de caractères pour affichage *)
  let string_of_propop2 = function
    | And -> "∧"
    | Or -> "∨"
    | Impl -> "=>" ;;
  let string_of_propop1 = function
    | Neg -> "¬" ;;
  let string_of_propop0 = function
    | True -> "true"
    | False -> "false" ;;
  let string_of_relop2 = function
    | LE -> "≤"
    | LT -> "<"
    | GE -> "≥"
    | GT -> ">"
    | Eq -> "="
    | NEq -> "≠" ;;
  let string_of_exprop2 = function
    | Plus -> "+"
    | Minus -> "−" ;;
end ;;



(*** Éléments structurels d'un BRN ***)

(*
module type BRN =
sig
  type var ;;
  type mult ;;
(** Fonctions utilitaires *)
  val varlist : var list ;;
  val varequals : var -> var -> bool ;;
  val multcompare : mult -> mult -> int ;;
(** Conversion des noms en chaînes de caractères pour ASP *)
  val asp_of_var : var -> string ;;
  val asp_of_mult : mult -> string ;;
(** Conversion des noms en chaînes de caractères pour affichage *)
  val string_of_var : var -> string ;;
  val string_of_mult : mult -> string ;;
end ;;
*)



(*
(** Exemple jouet *)
module ToyExample =
struct
  type var = A ;;
  type mult = M ;;

  let varlist = [A] ;;
  let varequals v1 v2 =
    match v1, v2 with
      | A, A -> true ;;

  let asp_of_var = function
    | A -> "A" ;;
  let asp_of_mult = function
    | M -> "M" ;;

  let string_of_var = function
    | A -> "a" ;;
  let string_of_mult = function
    | M -> "m" ;;

  let multcompare m1 m2 =
    compare (asp_of_mult m1) (asp_of_mult m2) ;;
end ;;
*)

(*
(** Exemple du papier *)
module SimpleExample =
struct
  type var = A | B | C ;;
  type mult = L | Lambda | Sigma ;;  (* TODO: Analyse des formules traitées pour générer une varlist automatique et minimale *)
                                     (* TODO bis: Même chose pour les paramètres *)
  let varlist = [A ; B ; C] ;;
  let varequals v1 v2 =
    match v1, v2 with
      | A, A -> true
      | B, B -> true
      | C, C -> true
      | _, _ -> false ;;

  let asp_of_var = function
    | A -> "A"
    | B -> "B"
    | C -> "C" ;;
  let asp_of_mult = function
    | L -> "L"
    | Lambda -> "Lambda"
    | Sigma -> "Sigma" ;;

  let string_of_var = function
    | A -> "a"
    | B -> "b"
    | C -> "c" ;;
  let string_of_mult = function
    | L -> "l"
    | Lambda -> "λ"
    | Sigma -> "σ" ;;

  let multcompare m1 m2 =
    compare (asp_of_mult m1) (asp_of_mult m2) ;;
end ;;
*)

(*
(** Phage lambda (depuis la thèse d'Adrien Richard) *)
module PhageLambda =
struct
  type var = CI | Cro | CII | N ;;
  type mult = M1 | M2 | M3 | M4 | M5 | M6 | M7 | M8 ;;

  let varlist = [CI ; Cro ; CII ; N] ;;
  let varequals v1 v2 =
    match v1, v2 with
      | CI, CI -> true
      | Cro, Cro -> true
      | CII, CII -> true
      | N, N -> true
      | _, _ -> false ;;

  let asp_of_var = function
    | CI -> "CI"
    | Cro -> "Cro"
    | CII -> "CII"
    | N -> "N" ;;
  let asp_of_mult = function
    | M1 -> "M1"
    | M2 -> "M2"
    | M3 -> "M3"
    | M4 -> "M4"
    | M5 -> "M5"
    | M6 -> "M6"
    | M7 -> "M7"
    | M8 -> "M8" ;;

  let string_of_var = function
    | CI -> "CI"
    | Cro -> "Cro"
    | CII -> "CII"
    | N -> "N" ;;
  let string_of_mult = function
    | M1 -> "m1"
    | M2 -> "m2"
    | M3 -> "m3"
    | M4 -> "m4"
    | M5 -> "m5"
    | M6 -> "m6"
    | M7 -> "m7"
    | M8 -> "m8" ;;

  let multcompare m1 m2 =
    compare (asp_of_mult m1) (asp_of_mult m2) ;;
end ;;
*)

(** Cycle circadien (depuis projet Circlock)  *)
module Circlock =
struct
  type var = PERCRY | PER1 | PER2 | CRY1 | CRY2 | Clock | RevErbAlpha | BMAL1N | BMAL1C ;;
  type mult = PC | PER1PER1 | PER1PER2 | PER2PER2 | CRY1CRY1 | CRY1CRY2 | CRY2CRY2 | ClockBMAL | CBR | Inhib | Acetylation | Context ;;

  let varlist = [PERCRY ; PER1 ; PER2 ; CRY1 ; CRY2 ; Clock ; RevErbAlpha ; BMAL1N ; BMAL1C] ;;
  let varequals v1 v2 =
    match v1, v2 with
      | PERCRY, PERCRY -> true
      | PER1, PER1 -> true
      | PER2, PER2 -> true
      | CRY1, CRY1 -> true
      | CRY2, CRY2 -> true
      | Clock, Clock -> true
      | RevErbAlpha, RevErbAlpha -> true
      | BMAL1N, BMAL1N -> true
      | BMAL1C, BMAL1C -> true
      | _, _ -> false ;;

  let asp_of_var = function
    | PERCRY -> "PERCRY"
    | PER1 -> "PER1"
    | PER2 -> "PER2"
    | CRY1 -> "CRY1"
    | CRY2 -> "CRY2"
    | Clock -> "Clock"
    | RevErbAlpha -> "RevErbAlpha"
    | BMAL1N -> "BMAL1N"
    | BMAL1C -> "BMAL1C" ;;
  let asp_of_mult = function
    | PC -> "PC"
    | PER1PER1 -> "PER1PER1"
    | PER1PER2 -> "PER1PER2"
    | PER2PER2 -> "PER2PER2"
    | CRY1CRY1 -> "CRY1CRY1"
    | CRY1CRY2 -> "CRY1CRY2"
    | CRY2CRY2 -> "CRY2CRY2"
    | ClockBMAL -> "ClockBMAL"
    | CBR -> "CBR"
    | Inhib -> "Inhib"
    | Acetylation -> "Acetylation"
    | Context -> "Context" ;;

  let string_of_var = function
    | PERCRY -> "PER-CRY (N)"
    | PER1 -> "PER1 (C)"
    | PER2 -> "PER2 (C)"
    | CRY1 -> "CRY1 (C)"
    | CRY2 -> "CRY2 (C)"
    | Clock -> "Clock (N)"
    | RevErbAlpha -> "RevErbα (N)"
    | BMAL1N -> "BMAL1 (N)"
    | BMAL1C -> "BMAL1 (C)" ;;
  let string_of_mult = function
    | PC -> "PC"
    | PER1PER1 -> "PER1-PER1"
    | PER1PER2 -> "PER1-PER2"
    | PER2PER2 -> "PER2-PER2"
    | CRY1CRY1 -> "CRY1-CRY1"
    | CRY1CRY2 -> "CRY1-CRY2"
    | CRY2CRY2 -> "CRY2-CRY2"
    | ClockBMAL -> "Clock-BMAL (N actif)"
    | CBR -> "CB-R"
    | Inhib -> "inhib"
    | Acetylation -> "Acetylation"
    | Context -> "Context" ;;

  let multcompare m1 m2 =
    compare (asp_of_mult m1) (asp_of_mult m2) ;;
end ;;



(** Sélection du BRN à considérer *)
(* module BRN = ToyExample ;; *)
(* module BRN = SimpleExample ;; *)
(* module BRN = PhageLambda ;; *)
module BRN = Circlock ;;



(**************************************************)
(*** Grammaire pour les formules de multiplexes ***)
(**************************************************)
module FOLmult =
(*functor (BRN : BRN) ->*)  (* TODO: À fonctoriser *)
struct
  open Operators ;;

(** Expression et formules de multiplexes *)
  type multexpr =
    | MExprBin of exprop2 * multexpr * multexpr
    | MExprConst of int ;;
  type multformula =
    | MPropConst of propop0
    | MPropUn of propop1 * multformula
    | MPropBin of propop2 * multformula * multformula
    | MRel of relop2 * multexpr * multexpr
    | MAtom of BRN.var * int
    | MMult of BRN.mult ;;

(** Conversion des expressions et formules en chaîne de caractères pour affichage *)
  let rec string_of_multexpr = function
    | MExprBin(op, e1, e2) -> ("(" ^ (string_of_multexpr e1) ^ (string_of_exprop2 op) ^ (string_of_multexpr e2) ^ ")")
    | MExprConst(i) -> string_of_int i ;;
  let rec string_of_multformula = function
    | MPropConst(op) -> string_of_propop0 op
    | MPropUn(op, f) -> ((string_of_propop1 op) ^ (string_of_multformula f))
    | MPropBin(op, f1, f2) -> ("(" ^ (string_of_multformula f1) ^ " " ^ (string_of_propop2 op) ^ " " ^ (string_of_multformula f2) ^ ")")
    | MRel(op, e1, e2) -> ("(" ^ (string_of_multexpr e1) ^ (string_of_relop2 op) ^ (string_of_multexpr e2) ^ ")")
    | MAtom(v, i) -> ((BRN.string_of_var v) ^ "_" ^ (string_of_int i))
    | MMult(m) -> BRN.string_of_mult m ;;
end ;;



(*****************************)
(*** Informations d'un BRN ***)
(*****************************)

(*TODO: trouver un moyen de mettre en œuvre une signature BRNinfo
(** Informations portées par le BRN *)
module type BRNinfo =
(*functor (BRN : BRN) ->*)
sig
  include BRN ;;
  val getbound : BRN.var -> int ;;
  val getformula : BRN.mult -> FOLmult(BRN).multformula ;;
  val getpredec : BRN.var -> BRN.mult list ;;
end ;;
*)



(*
(** Exemple jouet *)
module ToyExampleInfo =
struct
  open Operators ;;
  open ToyExample ;;
  open FOLmult ;;

  let getbound = function
    | A -> 1 ;;
  let getformula = function
    | M -> MPropUn(Neg, MAtom(A, 1)) ;;
  let getpredec = function
    | A -> [M] ;;
end ;;
*)

(*
(** Exemple du papier *)
module SimpleExampleInfo =
struct
  open Operators ;;
  open SimpleExample ;;
  open FOLmult ;;

  let getbound = function
    | A -> 1
    | B -> 1
    | C -> 1 ;;
  let getformula = function
    | L -> MAtom(A, 1)
    | Lambda -> MPropUn(Neg, MAtom(C, 1))
    | Sigma -> MAtom(A, 1) ;;
  let getpredec = function
    | A -> []
    | B -> [Lambda ; Sigma]
    | C -> [L] ;;
end ;;
*)

(*
(** Phage lambda (depuis la thèse d'Adrien Richard) *)
module PhageLambdaInfo =
struct
  open Operators ;;
  open PhageLambda ;;
  open FOLmult ;;

  let getbound = function
    | CI -> 2
    | Cro -> 3
    | CII -> 1
    | N -> 1 ;;
  let getformula = function
    | M1 -> MAtom(CI, 2)
    | M2 -> MPropUn(Neg, MAtom(Cro, 1))
    | M3 -> MPropUn(Neg, MAtom(CI, 2))
    | M4 -> MPropUn(Neg, MAtom(Cro, 3))
    | M5 -> MAtom(CI, 1)
    | M6 -> MPropUn(Neg, MAtom(CI, 1))
    | M7 -> MPropUn(Neg, MAtom(Cro, 2))
    | M8 -> MAtom(N, 1) ;;
  let getpredec = function
    | CI -> [M1 ; M2 ; M5]
    | Cro -> [M3 ; M4]
    | CII -> [M3 ; M4 ; M8]
    | N -> [M6 ; M7] ;;
end ;;
*)

(** Cycle circadien (depuis projet Circlock)  *)
module CirclockInfo =
struct
  open Operators ;;
  open Circlock ;;
  open FOLmult ;;

  let getbound : var -> int = fun _ -> 1 ;;
  let getformula = function
    | PC -> MPropBin(And,
              MPropBin(Or, MPropBin(Or, MMult(PER1PER1), MMult(PER1PER2)), MMult(PER2PER2)),
              MPropBin(Or, MPropBin(Or, MMult(CRY1CRY1), MMult(CRY1CRY2)), MMult(CRY2CRY2)))
    | PER1PER1 -> MAtom(PER1, 1)
    | PER1PER2 -> MPropBin(Or, MAtom(PER1, 1), MAtom(PER2, 1))
    | PER2PER2 -> MAtom(PER2, 1)
    | CRY1CRY1 -> MAtom(CRY1, 1)
    | CRY1CRY2 -> MPropBin(Or, MAtom(CRY1, 1), MAtom(CRY2, 1))
    | CRY2CRY2 -> MAtom(CRY2, 1)
    | ClockBMAL -> MPropBin(And, MPropBin(And, MAtom(Clock, 1), MAtom(BMAL1N, 1)), MPropUn(Neg, MAtom(PERCRY, 1)))
    | CBR -> MPropBin(And, MPropUn(Neg, MAtom(RevErbAlpha, 1)), MMult(ClockBMAL))
    | Inhib -> MPropUn(Neg, MAtom(RevErbAlpha, 1))
    | Acetylation -> MPropBin(And, MAtom(BMAL1C, 1), MAtom(Clock, 1))
    | Context -> MPropConst(True) ;;
  let getpredec = function
    | PERCRY -> [PC]
    | PER1 -> [ClockBMAL]
    | PER2 -> [ClockBMAL ; Context]
    | CRY1 -> [CBR]
    | CRY2 -> [ClockBMAL]
    | Clock -> []
    | RevErbAlpha -> [ClockBMAL]
    | BMAL1N -> [Inhib]
    | BMAL1C -> [Acetylation] ;;
end ;;



(** Sélection des informations du BRN à considérer *)
(* module BRNinfo = ToyExampleInfo ;; *)
(* module BRNinfo = SimpleExampleInfo ;; *)
(* module BRNinfo = PhageLambdaInfo ;; *)
module BRNinfo = CirclockInfo ;;



(**********************************************************)
(*** Grammaire pour les formules de la logique de Hoare ***)
(**********************************************************)
module FOL =
(*functor (BRN : BRN) ->*)  (* TODO: À fonctoriser *)
struct
  open Operators ;;

(** Expressions et formules pour les triplets de Hoare et les conditions de programme *)
  type expr =
    | ExprBin of exprop2 * expr * expr
    | ExprVar of BRN.var
    | ExprParam of BRN.var * (BRN.mult list)
(*  | ExprVal of string (* TODO: Variables extérieures au triplet de Hoare ? *) *)
    | ExprConst of int ;;
  type formula =
    | PropConst of propop0
    | PropUn of propop1 * formula
    | PropBin of propop2 * formula * formula
    | Rel of relop2 * expr * expr
(*  | Mult of BRN.mult  (* TODOx: À supprimer ? Conversion en amont *) *)
    | FreshState of formula ;;

(** Conversion d'une formule de multiplexe en formule FOL générale *)
  open FOLmult ;;
  let formula_of_matom v i = Rel(GE, ExprVar v, ExprConst i) ;;
  let rec expr_of_multexpr = function
    | MExprBin(op, e1, e2) -> ExprBin(op, expr_of_multexpr e1, expr_of_multexpr e2)
    | MExprConst(i) -> ExprConst(i) ;;
  let rec formula_of_multformula = function
    | MPropConst(op) -> PropConst(op)
    | MPropUn(op, f) -> PropUn(op, formula_of_multformula f)
    | MPropBin(op, f1, f2) -> PropBin(op, formula_of_multformula f1, formula_of_multformula f2)
    | MRel(op, e1, e2) -> Rel(op, expr_of_multexpr e1, expr_of_multexpr e2)
    | MAtom(v, i) -> formula_of_matom v i
    | MMult(m) -> formula_of_multformula (BRNinfo.getformula m) ;;  (* Mult m *)

(** Substitution d'une expression à une variable *)
  let replace f v e =
    let rec replace_expr ex v' e' =
      match ex with
        | ExprBin(op, e1, e2) -> ExprBin(op, replace_expr e1 v' e', replace_expr e2 v' e')
        | ExprVar(v) -> if (BRN.varequals v v') then e' else ExprVar(v)
        | ExprParam(v, l) -> ExprParam(v, l)
        | ExprConst(i) -> ExprConst(i) in
    let rec replace_formula tree v' e' =
      match tree with
        | PropConst(op) -> PropConst(op)
        | PropUn(op, f) -> PropUn(op, replace_formula f v' e')
        | PropBin(op, f1, f2) -> PropBin(op, replace_formula f1 v' e', replace_formula f2 v' e')
        | Rel(op, e1, e2) -> Rel(op, replace_expr e1 v' e', replace_expr e2 v' e')
(*      | Mult(m) -> replace_formula (formula_of_multformula (BRNinfo.getformula m)) v' e' *)
        | FreshState(f) -> FreshState(f) in
    replace_formula f v e ;;

(** Conversion d'une expression en chaîne de caractères pour ASP *)
(* let rec asp_of_expr s =
    let paramlist = List.fold_right (fun v l -> (powerset_sort (BRNinfo.getpredec v) BRN.multcompare) :: l) BRN.varlist [] in
  function
    | ExprBin(op, e1, e2) -> ("(" ^ (asp_of_expr e1) ^ (asp_of_exprop2 op) ^ (asp_of_expr e2) ^ ")")
    | ExprVar(v) -> BRN.asp_of_var v (* TODO: Énumération préalable des variables ? *)
    | ExprParam(v, l) -> ("K_" ^ BRN.asp_of_var v ^ "_" ^ (string_of_int (index (List.nth paramlist (index BRN.varlist v)) (List.sort BRN.multcompare l))))
    | ExprConst(i) -> string_of_int i ;; *)
  let rec asp_of_state_expr s =
    let paramlist = List.fold_right (fun v l -> (powerset_sort (BRNinfo.getpredec v) BRN.multcompare) :: l) BRN.varlist [] in
  function
    | ExprBin(op, e1, e2) -> ("(" ^ (asp_of_state_expr s e1) ^ (asp_of_exprop2 op) ^ (asp_of_state_expr s e2) ^ ")")
    | ExprVar(v) -> (if s = 0 then "" else "S" ^ (string_of_int s) ^ "X") ^ (BRN.asp_of_var v) (* TODO: Énumération préalable des variables ? *)
    | ExprParam(v, l) -> ("K_" ^ BRN.asp_of_var v ^ "_" ^ (string_of_int (index (List.nth paramlist (index BRN.varlist v)) (List.sort BRN.multcompare l))))
    | ExprConst(i) -> string_of_int i ;;
  let asp_of_expr = asp_of_state_expr 0 ;;

(** Conversion en chaîne de caractères pour affichage *)
  let rec string_of_expr = function
    | ExprBin(op, e1, e2) -> ("(" ^ (string_of_expr e1) ^ (string_of_exprop2 op) ^ (string_of_expr e2) ^ ")")
    | ExprVar(v) -> BRN.string_of_var v
    | ExprParam(v, l) -> ("k_" ^ BRN.string_of_var v ^ ",{" ^ (string_of_list BRN.string_of_mult ";" l) ^ "}")
    | ExprConst(i) -> string_of_int i ;;
  let rec string_of_formula = function
    | PropConst(op) -> string_of_propop0 op
    | PropUn(op, f) -> ((string_of_propop1 op) ^ (string_of_formula f))
    | PropBin(op, f1, f2) -> ("(" ^ (string_of_formula f1) ^ " " ^ (string_of_propop2 op) ^ " " ^ (string_of_formula f2) ^ ")")
    | Rel(op, e1, e2) -> ("(" ^ (string_of_expr e1) ^ (string_of_relop2 op) ^ (string_of_expr e2) ^ ")")
(*  | Mult(m) -> BRN.string_of_mult m *)
    | FreshState(f) -> ("∀ state σ, " ^ (string_of_formula f)) ;;
end ;;



(************************************************)
(*** Grammaire pour les programmes impératifs ***)
(************************************************)
module Prog =
(*functor (BRN : BRN) ->*)  (* TODO: À fonctoriser *)
struct
(*  module FOLmult = FOLmult(BRN) ;; *)
(*  module FOL = FOL(BRN) ;;*)

(** Grammaire des programmes impératifs *)
  type prog =
    | Skip
(*  | Set of BRN.var * int *)  (* TODO: Instruction Set (knockdown) ? *)
    | Incr of BRN.var
    | Decr of BRN.var
    | Seq of prog * prog
    | If of FOL.formula * prog * prog
    | While of FOL.formula * FOL.formula * prog
    | Forall of prog * prog
    | Exists of prog * prog ;;
(*  | Assert of FOL.formula *)  (* TODO: Instruction Assert ? *)

(** Conversion d'un programme impératif en chaîne de caractères pour affichage avec indentations *)
  let string_of_prog_indent p =
    let nli indent = ("\n" ^ (String.make indent ' ')) in
    let newindent = 2 in
    let rec string_of_prog_indentation p indent =
      (match p with
        | Skip -> "SKIP"
        | Incr(v) -> ((BRN.string_of_var v) ^ "++")
        | Decr(v) -> ((BRN.string_of_var v) ^ "−−")
        | Seq(p1, p2) -> ((string_of_prog_indentation p1 indent) ^ " ;"
          ^ (nli indent) ^ (string_of_prog_indentation p2 indent))
        | If(c, p1, p2) -> ("IF " ^ (FOL.string_of_formula c) ^ " THEN"
          ^ (nli (indent + newindent)) ^ (string_of_prog_indentation p1 (indent + newindent))
          ^ (nli indent) ^ "ELSE"
          ^ (nli (indent + newindent)) ^ (string_of_prog_indentation p2 (indent + newindent))
          ^ (nli indent) ^ "END IF")
        | While(c, i, p) -> ("WHILE " ^ (FOL.string_of_formula c) ^ " WITH " ^ (FOL.string_of_formula i) ^ " DO"
          ^ (nli (indent + newindent)) ^ (string_of_prog_indentation p (indent + newindent))
          ^ (nli indent) ^ "END WHILE")
        | Forall(p1, p2) -> ("∀(" ^ (string_of_prog_indentation p1 indent) ^ ", " ^ (string_of_prog_indentation p2 indent) ^ ")")
        | Exists(p1, p2) -> ("∃(" ^ (string_of_prog_indentation p1 indent) ^ ", " ^ (string_of_prog_indentation p2 indent) ^ ")")) in
  string_of_prog_indentation p 0 ;;

(** Conversion d'un programme impératif en chaîne de caractères pour affichage en ligne *)
  let rec string_of_prog = function
    | Skip -> "SKIP"
    | Incr(v) -> ((BRN.string_of_var v) ^ "++")
    | Decr(v) -> ((BRN.string_of_var v) ^ "−−")
    | Seq(p1, p2) -> ((string_of_prog p1) ^ " ; " ^ (string_of_prog p2))
    | If(c, p1, p2) -> ("(IF " ^ (FOL.string_of_formula c) ^ " THEN " ^ (string_of_prog p1) ^ " ELSE " ^ (string_of_prog p2) ^ " END IF)")
    | While(c, i, p) -> ("(WHILE " ^ (FOL.string_of_formula c) ^ " WITH " ^ (FOL.string_of_formula i) ^ " DO " ^ (string_of_prog p) ^ " END WHILE)")
    | Forall(p1, p2) -> ("∀(" ^ (string_of_prog p1) ^ ", " ^ (string_of_prog p2) ^ ")")
    | Exists(p1, p2) -> ("∃(" ^ (string_of_prog p1) ^ ", " ^ (string_of_prog p2) ^ ")") ;;
end ;;



(*******************************************************)
(*** Calcul de prédicats : plus faible pré-condition ***)
(*******************************************************)
module WP :
sig
  val wp : Prog.prog -> FOL.formula -> FOL.formula ;;
end =
(* functor (BRNinfo : BRNinfo) -> *)  (* TODO: À fonctoriser *)
struct
(*  module FOL = FOL(BRN) ;;*)
  open List ;;
  open Operators ;;
  open FOL ;;
  open Prog ;;

(** Formules de type PHI *)
  let phi v omega =
    let rec phi_eval v l omega =
      match l with
        | [] -> PropConst True (*raise (Invalid_argument "phi_eval: Empty l list for phi_eval")*)
        | h :: [] -> if (mem h omega)
                     then formula_of_multformula (BRNinfo.getformula h)  (* Mult h *)
                     else PropUn(Neg, formula_of_multformula (BRNinfo.getformula h))  (* Mult h *)
        | h :: t -> PropBin(And, (if (mem h omega)
                                  then formula_of_multformula (BRNinfo.getformula h)  (* Mult h *)
                                  else PropUn(Neg, formula_of_multformula (BRNinfo.getformula h))),  (* Mult h *)
                                 phi_eval v t omega) in
    phi_eval v (BRNinfo.getpredec v) omega ;;
  let rec conj_phi_param v l comp =
    match l with
      | [] -> raise (Invalid_argument "conj_phi_param: Empty powerset (should contain at least [])")
      | h :: [] -> PropBin(Impl, phi v h, Rel(comp, ExprParam(v, h), ExprVar v))
      | h :: t -> PropBin(And, PropBin(Impl, phi v h, Rel(comp, ExprParam(v, h), ExprVar v)), conj_phi_param v t comp) ;;
  let phi_incr v = PropBin(And, PropBin(And, Rel(LE, ExprConst 0, ExprVar v), Rel(LT, ExprVar v, ExprConst (BRNinfo.getbound v))),
                                conj_phi_param v (powerset (BRNinfo.getpredec v)) GT) ;;
  let phi_decr v = PropBin(And, PropBin(And, Rel(LT, ExprConst 0, ExprVar v), Rel(LE, ExprVar v, ExprConst (BRNinfo.getbound v))),
                                conj_phi_param v (powerset (BRNinfo.getpredec v)) LT) ;;

(** Plus faible pré-condition *)
  let incr_formula v = ExprBin(Plus, ExprVar v, ExprConst 1) ;;
  let decr_formula v = ExprBin(Minus, ExprVar v, ExprConst 1) ;;
  let rec wp prog post =
    match prog with
      | Skip -> post
      | Incr v -> PropBin(And, phi_incr v, replace post v (incr_formula v))
      | Decr v -> PropBin(And, phi_decr v, replace post v (decr_formula v))
      | Seq (p1, p2) -> wp p1 (wp p2 post)
      | If (c, p1, p2) -> PropBin(And, PropBin(Impl, c, wp p1 post),
                                       PropBin(Impl, PropUn(Neg, c), wp p2 post))
      | While (c, i, p) -> PropBin(And, i, PropBin(And, FreshState(PropBin(Impl, PropBin(And, i, c), wp p i)),
                                                        FreshState(PropBin(Impl, PropBin(And, i, PropUn(Neg, c)), post))))
      | Forall (p1, p2) -> PropBin(And, wp p1 post, wp p2 post)
      | Exists (p1, p2) -> PropBin(Or, wp p1 post, wp p2 post) ;;
end ;;



(*****************************************************************)
(*** Traduction d'une formule en programme ASP pour résolution ***)
(*****************************************************************)
module ASP =
struct
  open BRN ;;
  open BRNinfo ;;
  open Operators ;;
  open FOL ;;
  open Prog ;;
  open WP ;;

(** Traduction d'une formule en ASP *)
  let write_asp out f =
    let rec power x i = if i = 0 then 1 else x * power x (i - 1) in
    let rec from_up_to n m = if n > m then [] else n :: (from_up_to (n + 1) m) in
    let asp_of_state_var s v = (if s = 0 then "" else "S" ^ (string_of_int s) ^ "X") ^ (BRN.asp_of_var v) in
    let asp_of_state_param p = let (v, n) = p in "K_" ^ (asp_of_var v) ^ "_" ^ (string_of_int n) in
    let param_enum_var v = List.fold_right (fun n l -> (v, n) :: l) (from_up_to 0 ((power 2 (List.length (BRNinfo.getpredec v))) - 1)) [] in
    let param_enum = List.fold_right (fun v l -> (param_enum_var v) @ l) BRN.varlist [] in
(*  let vars_enum s = ((string_of_list (asp_of_state_var s) "," BRN.varlist) ^ "," ^ (string_of_list (asp_of_state_param s) "," param_enum)) in *)
(*  let hard_state_enum s = ((if s = 0 then "main_" else "") ^ "state(" ^ (vars_enum s) ^ ")") in
    let soft_state_enum s = if s = 0 then "" else ("state(" ^ (string_of_list (asp_of_state_var s) "," BRN.varlist) ^ ")") ^ ", " in *)
    let state_enum s = ((if s = 0 then "main_" else "") ^ "state(" ^ (string_of_list (asp_of_state_var s) "," BRN.varlist) ^ ")") in
    let soft_state_enum s = if s = 0 then "" else ((state_enum s) ^ ", ") in
    let params_enum = ("params(" ^ (string_of_list asp_of_state_param "," param_enum) ^ ")") in
    let iatom = ref (-1) in
    let nextatom s = iatom := !iatom + 1 ; "prop" ^ (string_of_int !iatom) ^
                                            (if s = 0 then "" else "(" ^ (string_of_list (asp_of_state_var s) "," BRN.varlist) ^ ")") in
    let istate = ref (0) in
    let nextstate () = istate := !istate + 1 ; !istate in
    let rec write_atom_state out f s =
      match f with
        | PropConst(op) -> (match op with
                              | True -> "t"
                              | False -> "f")
        | PropUn(op, f1) -> (let ncurrent = nextatom s in
                             let nf1 = write_atom_state out f1 s in
                               (match op with
                                  | Neg -> output_string out (ncurrent ^ " :- " ^ (soft_state_enum s) ^ "not " ^ nf1 ^ ".\n")
                               ) ; ncurrent)
        | PropBin(op, f1, f2) -> (let ncurrent = nextatom s in
                                  let nf1 = write_atom_state out f1 s in
                                  let nf2 = write_atom_state out f2 s in
                                    (match op with
                                       | And -> output_string out (ncurrent ^ " :- " ^ (soft_state_enum s) ^ nf1 ^ ", " ^ nf2 ^ ".\n")
                                       | Or -> output_string out (ncurrent ^ " :- " ^ (soft_state_enum s) ^ nf1 ^ ".\n" ^
                                                                  ncurrent ^ " :- " ^ (soft_state_enum s) ^ nf2 ^ ".\n")
                                       | Impl -> output_string out (ncurrent ^ " :- " ^ (soft_state_enum s) ^ nf2 ^ ".\n" ^
                                                                    ncurrent ^ " :- " ^ (soft_state_enum s) ^ "not " ^ nf1 ^ ".\n")
                                    ) ; ncurrent)
        | Rel(op, e1, e2) -> (let ncurrent = nextatom s in
                                output_string out (ncurrent ^ " :- " ^ (state_enum s) ^ ", " ^ params_enum ^ ", " ^
                                                     (asp_of_state_expr s e1) ^ " " ^ (asp_of_relop2 op) ^ " " ^ (asp_of_state_expr s e2) ^ ".\n")
                                ; ncurrent)
        | FreshState(f) -> (let ncurrent = nextatom s in
                            let nf1 = nextatom s in
                            let ns = nextstate () in
                            let nf2 = write_atom_state out f ns in
                              output_string out (ncurrent ^ " :- not " ^ nf1 ^ ".\n") ;
                              output_string out (nf1 ^ " :- " ^ (state_enum ns) ^ ", not " ^ nf2 ^ ".\n")
                              ; ncurrent) in
    let write_atoms out f =
      output_string out ("main_state :- " ^ write_atom_state out f 0 ^ ".\n") ;
      output_string out "\nt.\n:- f.\n\n:- not main_state.\n" in
    let write_header out =
      List.iter (fun v -> output_string out ("varbound_" ^ (asp_of_var v) ^ "(0.." ^ (string_of_int (getbound v)) ^ ").\n")) BRN.varlist ;
      output_string out "\n" ;
      List.iter (fun v -> output_string out ("1 {var_" ^ (asp_of_var v) ^ "(X) : varbound_" ^ (asp_of_var v) ^ "(X)} 1.\n")) BRN.varlist ;
      output_string out "\n" ;
      List.iter (fun v ->
        (List.iter
          (fun n -> output_string out ("1 {param_" ^ (asp_of_var v) ^ "_" ^ (string_of_int n) ^ "(X) : varbound_" ^ (asp_of_var v) ^ "(X)} 1.\n"))
          (from_up_to 0 ((power 2 (List.length (BRNinfo.getpredec v))) - 1))
        )) BRN.varlist ;
      output_string out "\n" ;

      output_string out ((state_enum 0) ^ " :- ") ;
      output_string out (string_of_list (fun v -> "var_" ^ (asp_of_var v) ^ "(" ^ (asp_of_var v) ^ ")") ", " BRN.varlist) ;
      output_string out ".\n\n" ;

      output_string out ("state(" ^ (string_of_list asp_of_var "," BRN.varlist) ^ ") :- ") ;
      output_string out (string_of_list (fun v -> "varbound_" ^ (asp_of_var v) ^ "(" ^ (asp_of_var v) ^ ")") ", " BRN.varlist) ;
      output_string out ".\n\n" ;

      output_string out (params_enum ^ " :- ") ;
      output_string out (string_of_list
        (fun c -> let (v, n) = c in "param_" ^ (asp_of_var v) ^ "_" ^ (string_of_int n) ^ "(K_" ^ (asp_of_var v) ^ "_" ^ (string_of_int n) ^ ")")
        ", " param_enum) ;
      output_string out ".\n\n" ;



(*    output_string out ((hard_state_enum 0) ^ " :- ") ;
      List.iter (fun v -> output_string out ("var_" ^ (asp_of_var v) ^ "(" ^ (asp_of_var v) ^ "), ")) BRN.varlist ;
      output_string out (string_of_list
        (fun c -> let (v, n) = c in "param_" ^ (asp_of_var v) ^ "_" ^ (string_of_int n) ^ "(K_" ^ (asp_of_var v) ^ "_" ^ (string_of_int n) ^ ")")
        ", " param_enum) ;
      output_string out ".\n\n" ; *)
(*    output_string out ("state(" ^ (string_of_list (asp_of_state_var 0) "," BRN.varlist) ^ ") :- ") ;
      output_string out (string_of_list (fun v -> "varbound_" ^ (asp_of_var v) ^ "(" ^ (asp_of_var v) ^ ")") ", " BRN.varlist) ;
(*    List.iter (fun v -> output_string out ("varbound_" ^ (asp_of_var v) ^ "(" ^ (asp_of_var v) ^ "), ")) BRN.varlist ; *)
(*    output_string out (string_of_list
        (fun c -> let (v, n) = c in "param_" ^ (asp_of_var v) ^ "_" ^ (string_of_int n) ^ "(K_" ^ (asp_of_var v) ^ "_" ^ (string_of_int n) ^ ")")
        ", " param_enum) ; *)
      output_string out ".\n\n" ; in *)
in
    let write_footer out =
      output_string out ("\n#hide.\n#show main_state/" ^ (string_of_int (List.length BRN.varlist)) ^
                         ".\n#show params/" ^ (string_of_int (List.length param_enum)) ^ ".\n\n") in
    write_header out ; write_atoms out f ; write_footer out ;;

(** Écriture dans un fichier *)
  let write_example f file =
    let out = open_out file in
  (*    write_asp stdout f ;  (* Affichage à l'écran *) *)
      write_asp out f ;
      close_out out ;;

(** Affichage des correspondances entre paramètres et variables ASP *)
  let asp_params () =
    let paramlist =
      List.fold_right (fun v l -> (powerset_sort (BRNinfo.getpredec v) BRN.multcompare) :: l) BRN.varlist [] in
    List.iter (fun v ->
      List.iter (fun lp -> print_endline ((string_of_expr (ExprParam(v, lp))) ^ " ..... " ^ (asp_of_expr (ExprParam(v, lp)))))
      (List.nth paramlist (index BRN.varlist v))) BRN.varlist ;;
(*    List.iter (fun n -> print_string ((string_of_expr (ExprParam(fst n, snd n))) ^ " ..... " ^ (asp_of_expr (ExprParam(fst n, snd n))))) paramlist ;; *)
end ;;



(*******************)
(*** Bac à sable ***)
(*******************)

open BRN ;;
open BRNinfo ;;

open Operators ;;
open FOL ;;
open Prog ;;
open WP ;;

open ASP ;;

(** Exemple jouet *)
(*
let prog1 = Seq(Incr A, Decr A) ;;
let post1 = Rel(Eq, ExprVar A, ExprConst 0) ;;
let pre1 = wp prog1 post1 ;;
print_endline "*********************************************************" ;;
print_endline "*** Exemple jouet : rétroaction négative en Incr/Decr ***" ;;
print_endline "*********************************************************" ;;
print_endline "*** Programme ***" ;;
print_endline (string_of_prog_indent prog1) ;;
print_endline "*** Post-condition ***" ;;
print_endline (string_of_formula post1) ;;
print_endline "*** Plus faible pré-condition calculée ***" ;;
print_endline (string_of_formula pre1) ;;
write_example pre1 "toyex-incrdecr.lp" ;;
(* OK : Une solution : A = 0, k_a,{} = 0, k_a,{m} = 1  *)

let prog2 = While(PropConst True, Rel(Eq, ExprVar A, ExprConst 0), Seq(Incr A, Decr A)) ;;
let post2 = PropConst True (* Rel(Eq, ExprVar A, ExprConst 0) *) ;;
let pre2 = wp prog2 post2 ;;
print_endline "*****************************************************" ;;
print_endline "*** Exemple jouet : rétroaction négative en While ***" ;;
print_endline "*****************************************************" ;;
print_endline "*** Programme ***" ;;
print_endline (string_of_prog_indent prog2) ;;
print_endline "*** Post-condition ***" ;;
print_endline (string_of_formula post2) ;;
print_endline "*** Plus faible pré-condition calculée ***" ;;
print_endline (string_of_formula pre2) ;;
write_example pre2 "toyex-while.lp" ;;
(* OK : Une solution : A = 0, k_a,{} = 0, k_a,{m} = 1  *)

let prog3 = If(Rel(Eq, ExprVar A, ExprConst 1),
  While(PropConst True, Rel(Eq, ExprVar A, ExprConst 1), Seq(Decr A, Incr A)),
  While(PropConst True, Rel(Eq, ExprVar A, ExprConst 0), Seq(Incr A, Decr A))) ;;
let post3 = PropConst True (* Rel(Eq, ExprVar A, ExprConst 0) *) ;;
let pre3 = wp prog3 post3 ;;
print_endline "********************************************" ;;
print_endline "*** Exemple jouet : rétroaction complexe ***" ;;
print_endline "********************************************" ;;
print_endline "*** Programme ***" ;;
print_endline (string_of_prog_indent prog3) ;;
print_endline "*** Post-condition ***" ;;
print_endline (string_of_formula post3) ;;
print_endline "*** Plus faible pré-condition calculée ***" ;;
print_endline (string_of_formula pre3) ;;
write_example pre3 "toyex-complex.lp" ;;
(* OK : Deux solutions : A = 0/1, k_a,{} = 0, k_a,{m} = 1  *)
*)



(** Exemple du papier *)
(*
let prog1 = Seq(Seq(Incr B, Incr C), Decr B) ;;
let post1 = Rel(Eq, ExprVar B, ExprConst 0) ;;
(* let pre1 = wp prog1 post1 ;; (* Sans raffinage *) *)
let pre1 = PropBin(And, wp prog1 post1, Rel(Eq, ExprVar A, ExprConst 1)) ;; (* Avec raffinage *)
print_endline "***************************" ;;
print_endline "*** Exemple 1 du papier ***" ;;
print_endline "***************************" ;;
print_endline "*** Programme ***" ;;
print_endline (string_of_prog_indent prog1) ;;
print_endline "*** Post-condition ***" ;;
print_endline (string_of_formula post1) ;;
print_endline "*** Plus faible pré-condition calculée ***" ;;
print_endline (string_of_formula pre1) ;;
write_example pre1 "simpleex1.lp" ;;
(* OK : 16 solutions (avec raffinage) *)
(* Contraintes trouvées : a = 1, b = 0, c = 0, k_b,{σ} = 0, k_b,{λ;σ} = 1, k_c,{l} = 1 *)

let prog2 = Seq(Incr B, Decr B) ;;
let post2 = post1 ;;
let pre2 = wp prog2 post2 ;;
print_endline "***************************" ;;
print_endline "*** Exemple 2 du papier ***" ;;
print_endline "***************************" ;;
print_endline "*** Programme ***" ;;
print_endline (string_of_prog_indent prog2) ;;
print_endline "*** Post-condition ***" ;;
print_endline (string_of_formula post2) ;;
print_endline "*** Plus faible pré-condition calculée ***" ;;
print_endline (string_of_formula pre2) ;;
write_example pre2 "simpleex2.lp" ;;
(* OK : Aucune solution *)

let prog3 = Incr C ;;
let post3 = post1 ;;
let pre3 = wp prog3 post3 ;;
print_endline "***************************" ;;
print_endline "*** Exemple 3 du papier ***" ;;
print_endline "***************************" ;;
print_endline "*** Programme ***" ;;
print_endline (string_of_prog_indent prog3) ;;
print_endline "*** Post-condition ***" ;;
print_endline (string_of_formula post3) ;;
print_endline "*** Plus faible pré-condition calculée ***" ;;
print_endline (string_of_formula pre3) ;;
write_example pre3 "simpleex3.lp" ;;
(* OK : 128 solutions *)

let prog4 = While(Rel(LT, ExprVar B, ExprConst 1), PropConst True, Exists(Exists(Exists(Incr B, Decr B), Incr C), Decr C)) ;;
let post4 = Rel(Eq, ExprVar B, ExprConst 1) ;;
(* let pre4 = wp prog4 post4 ;; (* Sans raffinage *) *)
let pre4 = PropBin(And, wp prog4 post4, PropBin(And, PropBin(And, PropBin(And, PropBin(And, PropBin(And, (* Avec raffinage *)
  Rel(Eq, ExprVar A, ExprConst 1), Rel(Eq, ExprVar B, ExprConst 0)), Rel(Eq, ExprVar C, ExprConst 0)),
  Rel(Eq, ExprParam(B, [Sigma]), ExprConst 0)), Rel(Eq, ExprParam(B, [Lambda ; Sigma]), ExprConst 0)), Rel(Eq, ExprParam(C, [L]), ExprConst 0))) ;;
print_endline "***************************************************" ;;
print_endline "*** Exemple 4 du papier : While avec I ≡ (True) ***" ;;
print_endline "***************************************************" ;;
print_endline "*** Programme ***" ;;
print_endline (string_of_prog_indent prog4) ;;
print_endline "*** Post-condition ***" ;;
print_endline (string_of_formula post4) ;;
print_endline "*** Plus faible pré-condition calculée ***" ;;
print_endline (string_of_formula pre4) ;;
write_example pre4 "simpleex4.lp" ;;
(* OK : Aucune solution (avec raffinage) *)

let prog4b = While(Rel(LT, ExprVar B, ExprConst 1), Rel(Eq, ExprVar A, ExprConst 0), Exists(Exists(Exists(Incr B, Decr B), Incr C), Decr C)) ;;
let post4b = Rel(Eq, ExprVar B, ExprConst 1) ;;
(* let pre4b = wp prog4b post4b ;; (* Sans raffinage *) *)
let pre4b = PropBin(And, wp prog4b post4b, PropBin(And, PropBin(And, PropBin(And, PropBin(And, PropBin(And, (* Avec raffinage *)
  Rel(Eq, ExprVar A, ExprConst 1), Rel(Eq, ExprVar B, ExprConst 0)), Rel(Eq, ExprVar C, ExprConst 0)),
  Rel(Eq, ExprParam(B, [Sigma]), ExprConst 0)), Rel(Eq, ExprParam(B, [Lambda ; Sigma]), ExprConst 0)), Rel(Eq, ExprParam(C, [L]), ExprConst 0))) ;;
print_endline "****************************************************" ;;
print_endline "*** Exemple 4 du papier : While avec I ≡ (a = 0) ***" ;;
print_endline "****************************************************" ;;
print_endline "*** Programme ***" ;;
print_endline (string_of_prog_indent prog4b) ;;
print_endline "*** Post-condition ***" ;;
print_endline (string_of_formula post4b) ;;
print_endline "*** Plus faible pré-condition calculée ***" ;;
print_endline (string_of_formula pre4b) ;;
write_example pre4b "simpleex4b.lp" ;;
(* OK : Aucune solution (avec raffinage) *)
*)


(** Phage lambda (depuis la thèse d'Adrien Richard) *)
(*
(*Qa = N+; CII+; CI+; CI+; N−; CII−
Ra ≡ (CI = 2 ∧ Cro = 0 ∧ CII = 0 ∧ N = 0)*)

let progpl1 = Seq(Seq(Seq(Seq(Seq(Incr N, Incr CII), Incr CI), Incr CI), Decr N), Decr CII) ;;
let postpl1 = PropBin(And, PropBin(And, PropBin(And, Rel(Eq, ExprVar CI, ExprConst 2), Rel(Eq, ExprVar Cro, ExprConst 0)),
                                                   Rel(Eq, ExprVar CII, ExprConst 0)), Rel(Eq, ExprVar N, ExprConst 0)) ;;
let prepl1 = wp progpl1 postpl1 ;;
print_endline "*** Exemple 1 du phage lambda ***" ;;
print_endline "* Programme *" ;;
print_endline (string_of_prog_indent progpl1) ;;
print_endline "* Post-condition *" ;;
print_endline (string_of_formula postpl1) ;;
print_endline "* Plus faible pré-condition calculée *" ;;
print_endline (string_of_formula prepl1) ;;
print_endline "***************************" ;;

write_example prepl1 "phagelex1.lp" ;;
(* Mauvais : explosion combinatoire => trop de résultats *)
*)



(** Cycle circadien (d'après projet Circlock) *)

let prog1 = Skip ;;
let post1 = PropConst(True) ;;
let pre1 = wp prog1 post1 ;;

print_endline "*** Cycle circadien ***" ;;
print_endline "* Programme *" ;;
print_endline (string_of_prog_indent prog1) ;;
print_endline "* Post-condition *" ;;
print_endline (string_of_formula post1) ;;
print_endline "* Plus faible pré-condition calculée *" ;;
print_endline (string_of_formula pre1) ;;
print_endline "***************************" ;;

write_example pre1 "circlock1.lp" ;;



print_endline "   *****" ;;
asp_params () ;;

